// MainMenuUI.cpp: implementation of the CMainMenuUI class.
//
//////////////////////////////////////////////////////////////////////

#include "MainMenuUI.h"
#include "customvertex.h"
#include "dxutil.h"
#include "d3dutil.h"
#include <DXErr8.h>


#define INPUT_USABLE_ANGLE  3.1415f / 3.0f
#define NUM_RING_PARTICLES  175
#define RING_SPEED          12000

PARTICLEVERTEX g_ParticleVertices[] =
{
    //     x,      y,     z,       color,     tu,    tv        
    {  -0.3f,  -0.3f,  0.0f,  0xffffffff,   0.0f,  1.0f}, 
    {  -0.3f,   0.3f,  0.0f,  0xffffffff,   0.0f,  0.0f},
    {   0.3f,  -0.3f,  0.0f,  0xffffffff,   1.0f,  1.0f},
    {   0.3f,   0.3f,  0.0f,  0xffffffff,   1.0f,  0.0f},
};

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CMainMenuUI::CMainMenuUI() : 
    m_pd3dDevice(NULL), m_pD3DXMeshSphere(NULL), m_vMenuPos(0.0f, 0.0f, 0.0f),
    m_fItemRot(0.0f), m_iSelectedItem(-1), m_fSelectItemRot(0.0f),
    m_pParticleVB(NULL), m_iCurTexture(0),
    m_dwMode(MENU_MINIMIZED), m_iAnimFrame(0)
{
}

CMainMenuUI::~CMainMenuUI()
{
    DeleteDeviceObjects();
}

void CMainMenuUI::OnMouseMove(const POINT& mousepos, const TransformMatrices* transmats)
{
    if(m_dwMode == MENU_MOUSEDOWN)
        m_dwMode = MENU_DRAGGING;

    if(m_dwMode == MENU_DRAGGING)
    {
        D3DXVECTOR3 vWorld;
        D3DXVECTOR3 vMouse = D3DXVECTOR3((float)mousepos.x, (float)mousepos.y, 0.8f);
        D3DXVec3Unproject(&vWorld, &vMouse, &transmats->viewport,
                          &transmats->matProj, &transmats->matView, &transmats->matWorld);

        m_vMenuPos.x = vWorld.x;
        m_vMenuPos.y = vWorld.y;
        m_vMenuPos.z = 0.0f;
    }
}

void CMainMenuUI::OnLButtonDown(const POINT& mousepos, const TransformMatrices* transmats)
{
    if(m_dwMode == MENU_MINTOMAX ||
       m_dwMode == MENU_MAXTOMIN ||
       m_dwMode == MENU_MOUSEDOWN ||
       m_dwMode == MENU_DRAGGING)
       return;

    D3DXVECTOR3 vScreen;
    if(m_dwMode == MENU_MINIMIZED)
    {
        D3DXVec3Project(&vScreen, &m_vMenuPos, &transmats->viewport,
                        &transmats->matProj, &transmats->matView, &transmats->matWorld);
        if(((mousepos.x - vScreen.x) * (mousepos.x - vScreen.x) + 
            (mousepos.y - vScreen.y) * (mousepos.y - vScreen.y)) < 100)
        {
            m_dwMode = MENU_MOUSEDOWN;
        }
        return;
    }

    D3DXMATRIX          matWorld1, matWorld2, matTemp;
    FLOAT               fRotAngle;
    TransformMatrices   itemTransMats;
    itemTransMats.viewport = transmats->viewport;
    itemTransMats.matProj  = transmats->matProj;
    itemTransMats.matView  = transmats->matView;
    for(int x = 0; x < m_vectItems.size(); ++x)
    {
        fRotAngle = m_fItemRot + 2 * 3.1415f * x / m_vectItems.size();
        
        if(fRotAngle > 2 * 3.1415f)
            fRotAngle -= 2 * 3.1415f;

        //only check the front 90 degrees
        if(fRotAngle < 3.1415f / 2.0f - INPUT_USABLE_ANGLE ||
           fRotAngle > 3.1415f / 2.0f + INPUT_USABLE_ANGLE)
            continue;

        D3DXMatrixTranslation(&matWorld1, m_vMenuPos.x, m_vMenuPos.y, m_vMenuPos.z);
        D3DXMatrixMultiply(&matTemp, &matWorld1, &transmats->matWorld);

        D3DXMatrixRotationZ(&matWorld1, 3.1415f / 4.0f);
        D3DXMatrixMultiply(&matWorld2, &matWorld1, &matTemp);

        D3DXMatrixRotationX(&matWorld1, 3.1415f / 16.0f);
        D3DXMatrixMultiply(&matTemp, &matWorld1, &matWorld2);

        D3DXMatrixRotationY(&matWorld2, fRotAngle);
        D3DXMatrixMultiply(&matWorld1, &matWorld2, &matTemp);
        
        D3DXMatrixTranslation(&matWorld2, 1.25f, 0.0f, 0.0f);
        D3DXMatrixMultiply(&itemTransMats.matWorld, &matWorld2, &matWorld1);
        
        if(m_vectItems[x]->IsInBounds(mousepos, &itemTransMats))
        {
            m_iSelectedItem = x;
            return;
        }
    }

    //either close the secondary items or minimize
    if(m_iSelectedItem >= 0)
    {
        m_iSelectedItem = -1;
    }
    else
    {
        m_dwMode        = MENU_MAXTOMIN;
        m_iAnimFrame    = 0;
    }
}

void CMainMenuUI::OnLButtonUp(const POINT& mousepos, const TransformMatrices* transmats)
{
    if(m_dwMode == MENU_MOUSEDOWN)
    {
        m_dwMode = MENU_MINTOMAX;
        m_iAnimFrame = 0;
        return;
    }
    else if(m_dwMode == MENU_DRAGGING)
    {
        m_dwMode = MENU_MINIMIZED;
        m_iAnimFrame = 0;
    }
}

HRESULT CMainMenuUI::Render(const D3DXMATRIX &matWorld)
{
    // Setup a material
    D3DMATERIAL8 mtrl;
    D3DUtil_InitMaterial( mtrl, 0.5f, 0.0f, 0.5f );
    m_pd3dDevice->SetMaterial( &mtrl );

    // translate the menu to the correct spot
    D3DXMATRIX matTemp;
    D3DXMatrixTranslation(&matTemp, m_vMenuPos.x, m_vMenuPos.y, m_vMenuPos.z);
    m_pd3dDevice->SetTransform(D3DTS_WORLD, &matTemp);

    //////////////////////////////////////////////////////////
    // Render the main sphere mesh
    //////////////////////////////////////////////////////////
    
    switch(m_dwMode)
    {
    case MENU_DRAGGING:
    case MENU_MOUSEDOWN:
    case MENU_MINIMIZED:
        D3DXMatrixScaling(&matTemp, 0.1f, 0.1f, 0.1f);
        m_pd3dDevice->MultiplyTransform(D3DTS_WORLD, &matTemp);
        m_pD3DXMeshSphere->DrawSubset(0);
        return S_OK;
        break;
    case MENU_MINTOMAX:     //grow the sphere to the size of the maximized menu
        D3DXMatrixScaling(&matTemp, 0.1f + m_iAnimFrame / 100.0f, 
                                    0.1f + m_iAnimFrame / 100.0f,
                                    0.1f + m_iAnimFrame / 100.0f);
        m_pd3dDevice->MultiplyTransform(D3DTS_WORLD, &matTemp);
        m_pD3DXMeshSphere->DrawSubset(0);
        return S_OK;
        break;
    case MENU_MAXTOMIN:
        D3DXMatrixScaling(&matTemp, 1.0f - m_iAnimFrame / 100.0f, 
                                    1.0f - m_iAnimFrame / 100.0f,
                                    1.0f - m_iAnimFrame / 100.0f);
        m_pd3dDevice->MultiplyTransform(D3DTS_WORLD, &matTemp);
        m_pD3DXMeshSphere->DrawSubset(0);
        return S_OK;
        break;
    }
    
    m_pD3DXMeshSphere->DrawSubset(0);

    //Disable lighting and zwrite, Enable alpha and texture
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,           FALSE );
    m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE,   TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,           D3DBLEND_SRCALPHA );
    m_pd3dDevice->SetRenderState( D3DRS_DESTBLEND,          D3DBLEND_INVSRCALPHA );
    m_pd3dDevice->SetRenderState( D3DRS_ALPHATESTENABLE,    TRUE );
    m_pd3dDevice->SetRenderState( D3DRS_ALPHAREF,           0x08 );
    m_pd3dDevice->SetRenderState( D3DRS_ALPHAFUNC,          D3DCMP_GREATEREQUAL );

    //////////////////////////////////////////////////////////
    // Render the particle system
    //////////////////////////////////////////////////////////
    D3DXMATRIX matWorld1, matWorld2;
    
    D3DXMatrixTranslation(&matWorld1, m_vMenuPos.x, m_vMenuPos.y, m_vMenuPos.z);
    D3DXMatrixMultiply(&matTemp, &matWorld1, &matWorld);

    D3DXMatrixRotationZ(&matWorld2, 3.1415f / 4.0f);
    D3DXMatrixMultiply(&matWorld1, &matWorld2, &matTemp);

    D3DXMatrixRotationX(&matWorld2, 3.1415f / 16.0f);
    D3DXMatrixMultiply(&matTemp, &matWorld2, &matWorld1);

    D3DXMatrixRotationY(&matWorld2, m_fItemRot);
    D3DXMatrixMultiply(&matWorld1, &matWorld2, &matTemp);

    m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
    RenderParticleRing(matWorld1, FALSE);
    m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );

    //////////////////////////////////////////////////////////
    // Render the UI Items
    //////////////////////////////////////////////////////////
    FLOAT      fRotAngle;
    D3DXMATRIX matBillboard;
    
    for(int x = 0; x < m_vectItems.size(); ++x)
    {
        fRotAngle = 2 * 3.1415f * x / m_vectItems.size();
        D3DXMatrixRotationY(&matTemp, fRotAngle);
        D3DXMatrixMultiply(&matWorld2, &matTemp, &matWorld1);
        D3DXMatrixTranslation(&matTemp, 1.2f, 0.0f, 0.0f);
        D3DXMatrixMultiply(&matBillboard, &matTemp, &matWorld2);
        m_vectItems[x]->Render(matBillboard);
    }

    //////////////////////////////////////////////////////////
    // Render the secondary UI items
    //////////////////////////////////////////////////////////
    if(m_iSelectedItem >= 0)
    {
        fRotAngle = 2 * 3.1415f * m_iSelectedItem / m_vectItems.size();
        D3DXMatrixRotationY(&matTemp, fRotAngle);
        D3DXMatrixMultiply(&matWorld2, &matTemp, &matWorld1);
        D3DXMatrixRotationZ(&matTemp, m_fSelectItemRot);
        D3DXMatrixMultiply(&matWorld1, &matTemp, &matWorld2);
        
        m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
        RenderParticleRing(matWorld1, TRUE);
        m_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );

        for(int x = 0; x < m_vectItems.size(); ++x)
        {
            fRotAngle = 2 * 3.1415f * x / 10.0f;
            D3DXMatrixRotationZ(&matTemp, fRotAngle);
            D3DXMatrixMultiply(&matWorld2, &matTemp, &matWorld1);

            D3DXMatrixTranslation(&matTemp, 0.0f, 1.3f, 0.0f);
            D3DXMatrixMultiply(&matBillboard, &matTemp, &matWorld2);

            m_vectItems[x]->Render(matBillboard);
        }
    }

    //Enable lighting and zwrite, Disable alpha and texture
    m_pd3dDevice->SetTexture( 0, NULL );
    m_pd3dDevice->SetRenderState( D3DRS_ALPHATESTENABLE,    FALSE );
    m_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE,   FALSE );
    m_pd3dDevice->SetRenderState( D3DRS_LIGHTING,           TRUE );

    return S_OK;
}

HRESULT CMainMenuUI::FrameMove(const POINT& mousepos, const TransformMatrices* transmats)
{
    //move direction if within upper-right or lower-left quadrants
    //move in proportion to the distance from the center of the sphere
    //        |    |ROT
    //        |    |LFT
    //    ----+----+----
    //        |    |
    //        |    |
    //    ----+----+----
    //     ROT|    |
    //     RGT|    |

    static FLOAT fLastTime = DXUtil_Timer( TIMER_GETAPPTIME );
    FLOAT fTime = DXUtil_Timer( TIMER_GETAPPTIME );

    //if more than 20 frames skipped don't move and
    //make up for lost time if more than 20 frames skipped
    if(fTime - fLastTime >= 0.20f)
    {  
        fLastTime = fTime;
        return S_OK;
    }

    //no animation when minimized
    switch(m_dwMode)
    {
    case MENU_MINIMIZED:
        return S_OK;
        break;
    case MENU_MINTOMAX:     //grow the sphere to the size of the maximized menu
        while(fTime - fLastTime > 0.01f)
        {
            fLastTime += 0.01f;
            m_iAnimFrame += 2;
            if(m_iAnimFrame > 90)
                m_dwMode = MENU_MAXIMIZED;
        }
        return S_OK;
        break;
    case MENU_MAXTOMIN:     //shrink the sphere to the size of the minimized menu
        while(fTime - fLastTime > 0.01f)
        {
            fLastTime += 0.01f;
            ++m_iAnimFrame += 2;
            if(m_iAnimFrame > 90)
                m_dwMode = MENU_MINIMIZED;
        }
        return S_OK;
        break;
    }

    D3DXVECTOR3 vDelta;
    D3DXVec3Project(&vDelta, &m_vMenuPos, &transmats->viewport,
                    &transmats->matProj, &transmats->matView, &transmats->matWorld);
    vDelta.x = mousepos.x - vDelta.x;
    vDelta.y = mousepos.y - vDelta.y;
    
    //loop until time is caught up
    while(fTime - fLastTime > 0.01f)
    {
        fLastTime += 0.01f;

        if(m_iSelectedItem < 0)
        {
            if(vDelta.x > 10.0f && vDelta.y < -10.0f)
                m_fItemRot += (float)sqrt(vDelta.x * vDelta.x + vDelta.y * vDelta.y) / RING_SPEED;
            else if(vDelta.x < -10.0f && vDelta.y > 10.0f)
                m_fItemRot -= (float)sqrt(vDelta.x * vDelta.x + vDelta.y * vDelta.y) / RING_SPEED;

            //keep rotation in bounds between 0 and 2pi
            while(m_fItemRot > 2 * 3.1415f)
                m_fItemRot -= 2 * 3.1415f;
            while(m_fItemRot < 0.0f)
                m_fItemRot += 2 * 3.1415f;
        }
        else
        {
            if(vDelta.x < -10.0f && vDelta.y < -10.0f)
                m_fSelectItemRot -= (float)sqrt(vDelta.x * vDelta.x + vDelta.y * vDelta.y) / RING_SPEED;
            else if(vDelta.x > 10.0f && vDelta.y > 10.0f)
                m_fSelectItemRot += (float)sqrt(vDelta.x * vDelta.x + vDelta.y * vDelta.y) / RING_SPEED;

            //keep rotation in bounds between 0 and 2pi
            while(m_fItemRot > 2 * 3.1415f)
                m_fSelectItemRot -= 2 * 3.1415f;
            while(m_fItemRot < 0.0f)
                m_fSelectItemRot += 2 * 3.1415f;
        }
    }

    static FLOAT fLastTextureTime = DXUtil_Timer( TIMER_GETAPPTIME );

    //rotate the textures around
    if(fTime - fLastTextureTime > 0.5f)
    {
        fLastTextureTime = fTime;
        ++m_iCurTexture;
        m_iCurTexture %= NUM_PART_TEXTURES;
    }

    return S_OK;
}

HRESULT CMainMenuUI::InitDeviceObjects(LPDIRECT3DDEVICE8 pd3dDevice)
{
    m_pd3dDevice = pd3dDevice;

    for(int j = 0; j < NUM_PART_TEXTURES; ++j)
    {
        m_pParticleTextures[j] = NULL;
    }
    
    //create the particle ring
    m_pParticles = new D3DVECTOR[NUM_RING_PARTICLES];
    FLOAT fAngle;
    for(int i = 0; i < NUM_RING_PARTICLES; ++i)
    {
        fAngle = rand() * 2 * 3.1415f / RAND_MAX;
        m_pParticles[i].x = (float)(1.25f * sin(fAngle));
        m_pParticles[i].y = (float)(1.25f * cos(fAngle));
        m_pParticles[i].z = rand() * 0.2f / RAND_MAX - 0.1f;
    }

    //create ten items
    for(int k = 0; k < 10; ++k)
        m_vectItems.push_back(new CMenuItem());

    HRESULT hr;
    // Create a sphere mesh using D3DX
    if( FAILED( hr = D3DXCreateSphere( m_pd3dDevice, 1.0f, 20, 10, &m_pD3DXMeshSphere, NULL ) ) )
        return DXTRACE_ERR_NOMSGBOX( "D3DXCreateSphere", hr );

    // Create a quad for rendering each particle
    if( FAILED( m_pd3dDevice->CreateVertexBuffer( sizeof(g_ParticleVertices),
                                                  D3DUSAGE_WRITEONLY, D3DFVF_PARTICLEVERTEX,
                                                  D3DPOOL_MANAGED, &m_pParticleVB ) ) )
    {
        return E_FAIL;
    }

    // Copy particle mesh data into vertexbuffer
    PARTICLEVERTEX* vBuf;
    m_pParticleVB->Lock( 0, 0, (BYTE**)&vBuf, 0 );
    memcpy(vBuf, g_ParticleVertices, sizeof(g_ParticleVertices)); 
    m_pParticleVB->Unlock();

    char pszBuf[20];
    strcpy(pszBuf, "flame_.tga");

    // Create the particle texture
    for(int l = 0; l < NUM_PART_TEXTURES; ++l)
    {
        pszBuf[5] = '0' + l;
        if( FAILED( D3DUtil_CreateTexture( m_pd3dDevice, pszBuf, &m_pParticleTextures[l] ) ) )
            return E_FAIL;
    }

    vector<CMenuItem*>::iterator it;
    for(it = m_vectItems.begin(); it != m_vectItems.end(); ++it)
        (*it)->InitDeviceObjects(m_pd3dDevice);

    return S_OK;
}

HRESULT CMainMenuUI::RestoreDeviceObjects()
{
    vector<CMenuItem*>::iterator it;
    for(it = m_vectItems.begin(); it != m_vectItems.end(); ++it)
        (*it)->RestoreDeviceObjects();
    return S_OK;
}

HRESULT CMainMenuUI::InvalidateDeviceObjects()
{
    vector<CMenuItem*>::iterator it;
    for(it = m_vectItems.begin(); it != m_vectItems.end(); ++it)
        (*it)->InvalidateDeviceObjects();
    return S_OK;
}

HRESULT CMainMenuUI::DeleteDeviceObjects()
{
    SAFE_RELEASE( m_pD3DXMeshSphere );
    SAFE_RELEASE( m_pParticleVB );

    for(int i = 0; i < NUM_PART_TEXTURES; ++i)
    {
        SAFE_RELEASE( m_pParticleTextures[i] );
    }
    
    SAFE_DELETE_ARRAY( m_pParticles );

    //delete the items
    vector<CMenuItem*>::iterator it;
    for(it = m_vectItems.begin(); it != m_vectItems.end(); ++it)
        SAFE_DELETE( *it );
    m_vectItems.resize(0);

    return S_OK;
}

//renders the ring of particles vertically or horizontally
//depending on the flag
void CMainMenuUI::RenderParticleRing(const D3DXMATRIX &matWorld, BOOL bVert)
{
    m_pd3dDevice->SetTexture( 0, m_pParticleTextures[m_iCurTexture] );

    // Loop through and render all particles
    m_pd3dDevice->SetStreamSource( 0, m_pParticleVB, sizeof(PARTICLEVERTEX) );
    m_pd3dDevice->SetVertexShader( D3DFVF_PARTICLEVERTEX );
    D3DXMATRIX matTemp, matBillboard;
    
    for(int i = 0; i < NUM_RING_PARTICLES; ++i)
    {
        if(bVert)
        {
            D3DXMatrixTranslation(&matTemp, m_pParticles[i].x,
                                  m_pParticles[i].y, m_pParticles[i].z);
        }
        else
        {
            D3DXMatrixTranslation(&matTemp, m_pParticles[i].x,
                                  m_pParticles[i].z, m_pParticles[i].y);
        }
        D3DXMatrixMultiply(&matBillboard, &matTemp, &matWorld);

        matBillboard._11 = matBillboard._22 = matBillboard._33 = 1.0f;
        matBillboard._12 = matBillboard._13 = matBillboard._21 = 0.0f;
        matBillboard._23 = matBillboard._31 = matBillboard._32 = 0.0f;

        m_pd3dDevice->SetTransform(D3DTS_WORLD, &matBillboard);

        // Render the billboard
        m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );
    }
}